unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, SQLite3Conn, SQLDB, Forms, Controls, Graphics, Dialogs,
  StdCtrls, ExtCtrls;

type

  { TForm1 }

  TForm1 = class(TForm)
    btnPasswords: TButton;
    btnCookies: TButton;
    Memo1: TMemo;
    rgBrowser: TRadioGroup;
    SQLite3Connection1: TSQLite3Connection;
    SQLQuery1: TSQLQuery;
    SQLTransaction1: TSQLTransaction;
    procedure btnCookiesClick(Sender: TObject);
    procedure btnPasswordsClick(Sender: TObject);
    procedure Memo1Change(Sender: TObject);
  private

  public
    procedure ExtractPasswords;
    procedure ExtractCookies;

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }
uses Windows, FileUtil, Base64, fpjson, jsonparser;

function DecodeBase64(const S: string; Strict: boolean = False): TBytes;
var
  Instream: TStringStream;
  Outstream: TMemoryStream;
  Decoder: TBase64DecodingStream;
begin
  Instream := TStringStream.Create(S);
  try
    Outstream := TMemoryStream.Create;
    try
      if Strict then
        Decoder := TBase64DecodingStream.Create(Instream, bdmStrict)
      else
        Decoder := TBase64DecodingStream.Create(Instream, bdmMIME);
      try
        Outstream.CopyFrom(Decoder, Decoder.Size);
        SetLength(Result, Outstream.Size);
        Outstream.Position := 0;
        Move(Outstream.Memory^, Result[0], Outstream.Size);
      finally
        Decoder.Free;
      end;
    finally
      Outstream.Free;
    end;
  finally
    Instream.Free;
  end;
end;

type
  TBlobData = record
    cbData: DWORD;
    pbData: LPBYTE;
  end;
  DATA_BLOB = TBlobData;
  PBlobData = ^TBlobData;
  LPLPWSTR = ^LPWSTR;

function CryptUnprotectData(pDataIn: PBlobData; ppszDataDescr: LPLPWSTR; pOptionalEntropy: PBlobData; pvReserved: Pointer; pPromptStruct: Pointer; dwFlags: DWORD; pDataOut: PBlobData): BOOL; stdcall;
  external 'crypt32.dll';

function GetEncryptedKey(BrowserIndex: Integer): TBytes;
var
  Content: string;
  ALocal, AFileName: string;
  P: TJSONParser;
  J, D: TJSONData;
  LocalStateEncodedKey: ansistring;
  Bytes: TBytes;
var
  DataIn: DATA_BLOB;
  DataOut: DATA_BLOB;
begin

  ALocal := SysUtils.GetEnvironmentVariable('LOCALAPPDATA');
  if ALocal = '' then exit;
  case BrowserIndex of
    1: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'Google\Chrome\User Data\Local State';
    0: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'BraveSoftware\Brave-Browser\User Data\Local State';
    else raise Exception.Create('Onbekende browser');
  end;
  if not FileExists(AFileName) then raise Exception.Create('Local State niet gevonden');

  Content := ReadFileToString(AFilename); // or TFile.ReadAllText from IOUtils for Delphi
  LocalStateEncodedKey := '';
  P := TJSONParser.Create(Content);
  try
    J := P.Parse;
    if Assigned(J) then
    begin
      D := J.FindPath('os_crypt');
      if Assigned(D) then
      begin
        D := D.FindPath('encrypted_key');
        if Assigned(D) then
          LocalStateEncodedKey := D.AsString;
      end;
    end;
    if LocalStateEncodedKey <> '' then
    begin
      Bytes := DecodeBase64(LocalStateEncodedKey, True);
      DataIn.pbData := @Bytes[5];
      DataIn.cbData := Length(Bytes) - 5;
      if CryptUnprotectData(@DataIn, nil, nil, nil, nil, 0, @DataOut) then
      begin
        SetLength(Result, DataOut.cbData);
        Move(DataOut.pbData^, Result[0], DataOut.cbData);
        LocalFree(cardinal(DataOut.pbData));
        //LocalFree(cardinal(DataIn.pbData)); // TBytes = managed
      end;
    end;
  finally
    if Assigned(J) then
      J.Free;
    P.Free;
  end;
end;

type
  BCRYPT_ALG_HANDLE = THandle;
  BCRYPT_HANDLE = THandle;
  BCRYPT_KEY_HANDLE = THandle;
  NTSTATUS = LONG;
  TNTStatus = NTSTATUS;

type
  _BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO = record
    cbSize: ULONG;
    dwInfoVersion: ULONG;
    pbNonce: PBYTE; //PUCHAR;
    cbNonce: ULONG;
    pbAuthData: PBYTE; //PUCHAR;
    cbAuthData: ULONG;
    pbTag: PBYTE; //PUCHAR;
    cbTag: ULONG;
    pbMacContext: PBYTE; //PUCHAR;
    cbMacContext: ULONG;
    cbAAD: ULONG;
    cbData: ULONGLONG;
    dwFlags: ULONG;
  end;
  BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO = _BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO;

const
  bcryptdll = 'bcrypt.dll';
  BCRYPT_CHAINING_MODE_: WideString = 'ChainingMode';
  BCRYPT_CHAIN_MODE_GCM_: WideString = 'ChainingModeGCM';

function BCryptOpenAlgorithmProvider(out phAlgorithm: BCRYPT_ALG_HANDLE; pszAlgId, pszImplementation: LPCWSTR; dwFlags: ULONG): TNTStatus; stdcall; external bcryptdll;
function BCryptSetProperty(hObject: BCRYPT_HANDLE; pszProperty: LPCWSTR; pbInput: PUCHAR; cbInput: ULONG; dwFlags: ULONG): TNTStatus; stdcall; external bcryptdll;
function BCryptGenerateSymmetricKey(hAlgorithm: BCRYPT_ALG_HANDLE; out phKey: BCRYPT_KEY_HANDLE; pbKeyObject: PUCHAR; cbKeyObject: ULONG; pbSecret: PUCHAR; cbSecret, dwFlags: ULONG): TNTStatus; stdcall; external bcryptdll;
function BCryptDecrypt(hKey: BCRYPT_KEY_HANDLE; pbInput: PUCHAR; cbInput: ULONG; pPaddingInfo: Pointer; pbIV: PUCHAR; cbIV: ULONG; pbOutput: PUCHAR; cbOutput: ULONG; out pcbResult: ULONG; dwFlags: ULONG): TNTStatus; stdcall; external bcryptdll;

function BDecrypt_GCM(Algoritme: lpcwstr; Encryped: array of byte; Output: pointer; const gKey, IV: array of byte): ULONG;
const
  AES_BLOCK_SIZE = 16; // tevens tag size
var
  hProvider: BCRYPT_ALG_HANDLE = 0;
  Decrypted: array[0..1023] of byte;
  hKey: BCRYPT_KEY_HANDLE = 0;
  Status: NTSTATUS;
  Info: BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO;
begin
  Result := 0;
  Status := BCryptOpenAlgorithmProvider(hProvider, Algoritme, nil, 0);
  if Status <> 0 then exit;
  Status := BCryptSetProperty(hProvider, pwidechar(BCRYPT_CHAINING_MODE_), @BCRYPT_CHAIN_MODE_GCM_[1], sizeof(BCRYPT_CHAIN_MODE_GCM_), 0);
  if Status <> 0 then exit;
  Status := BCryptGenerateSymmetricKey(hProvider, hKey, nil, 0, @gKey[0], sizeof(gKey), 0);
  if Status <> 0 then exit;
  FillChar(Decrypted, length(Decrypted), 0);
  FillChar(Info, sizeof(Info), 0);
  Info.cbSize := sizeof(Info);
  Info.dwInfoVersion := 1;
  Info.pbNonce := @IV[0]; // iv = nonce
  Info.cbNonce := SizeOf(IV);
  Info.pbTag := @Encryped[SizeOf(Encryped) - AES_BLOCK_SIZE]; // tag size
  Info.cbTag := AES_BLOCK_SIZE;
  Status := BCryptDecrypt(hkey, @Encryped[0], SizeOf(Encryped) - AES_BLOCK_SIZE, @Info, nil, 0, @Decrypted[0], length(Decrypted), Result, 0);
  if Status <> 0 then exit;
  if Output = nil then Output := AllocMem(Result);
  CopyMemory(Output, @decrypted[0], Result);
end;

procedure TForm1.btnPasswordsClick(Sender: TObject);
begin
  Memo1.Clear;
  ExtractPasswords;
end;

procedure TForm1.Memo1Change(Sender: TObject);
begin

end;

procedure TForm1.btnCookiesClick(Sender: TObject);
begin
  Memo1.Clear;
  ExtractCookies;
end;

procedure TForm1.ExtractPasswords;
var
  ALocal, AFileName: string;
  Encrypted_Value: TBytes;
  Key: TBytes;
  DataIn: DATA_BLOB;
  DataOut: DATA_BLOB;
  Line: string;
  Nonce: TBytes;
  InBuf: TBytes;
  OutBuf: TBytes;
begin
  Key := GetEncryptedKey(rgBrowser.ItemIndex);
  Memo1.Clear;
  Memo1.ScrollBars := ssVertical;
  Memo1.Lines.Add('Key = ' + TEncoding.ANSI.GetString(Key));

  ALocal := SysUtils.GetEnvironmentVariable('LOCALAPPDATA');
  if ALocal = '' then exit;
  case rgBrowser.ItemIndex of
    0: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'Google\Chrome\User Data\Default\Login Data';
    1: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'BraveSoftware\Brave-Browser\User Data\Default\Login Data';
    else raise Exception.Create('Onbekende browser');
  end;
  if not FileExists(AFileName) then raise Exception.Create('Login Data niet gevonden');

  try
    SQLite3Connection1.Databasename := AFilename;
    SQLQuery1.SQL.Text := 'SELECT origin_url, username_value, password_value FROM logins';
    SQLQuery1.Open;
  except
    on E: Exception do
    begin
      SQLQuery1.Close;
      SQLite3Connection1.Close;
      ShowMessage('Error opening database' + #13 + E.Message);
      exit;
    end;
  end;

  while not SQLQuery1.EOF do
  begin
    Line := '';
    Line := SQLQuery1.FieldByName('origin_url').AsString;
    Line := Line + ' ' + SQLQuery1.FieldByName('username_value').AsString;

    Encrypted_Value := SQLQuery1.FieldByName('password_value').AsBytes;
    // v10 (i.e. 0x763130)
    // of
    // DPAPI = 0x01000000D08C9DDF0115D1118C7A00C04FC297EB

    if Copy(TEncoding.ANSI.GetString(Encrypted_Value), 1, 3) = 'v10' then
    begin
      // v10 (i.e. 0x763130)
      // data = bytes.fromhex('763130...') # the encrypted cookie
      // nonce = data[3:3+12]
      // ciphertext = data[3+12:-16]
      // tag = data[-16:]
      // cipher = AES.new(decrypted_key, AES.MODE_GCM, nonce=nonce)
      // plaintext = cipher.decrypt_and_verify(ciphertext, tag) # the decrypted cookie

      if Length(Encrypted_Value) > 3 + 12 + 16 then
      begin
        SetLength(Nonce, 12);
        SetLength(InBuf, Length(Encrypted_Value) - 15);
        SetLength(OutBuf, Length(InBuf) - 16 { tag});
        Move(Encrypted_Value[3], Nonce[0], 12);
        Move(Encrypted_Value[15], InBuf[0], Length(InBuf));
        if BDecrypt_GCM('AES', InBuf, @OutBuf[0], Key, Nonce) <> 0 then
        begin
          Line := Line + ' - ' + TEncoding.UTF8.GetString(OutBuf);
        end;
      end;

    end
    else
    begin
      // DPAPI = 0x01000000D08C9DDF0115D1118C7A00C04FC297EB

      DataIn.pbData := @Encrypted_Value[0];
      DataIn.cbData := Length(Encrypted_Value);
      if CryptUnprotectData(@DataIn, nil, nil, nil, nil, 0, @DataOut) then
      begin
        SetLength(OutBuf, DataOut.cbData);
        Move(DataOut.pbData^, OutBuf[0], DataOut.cbData);
        LocalFree(cardinal(DataOut.pbData));
        // LocalFree(cardinal(DataIn.pbData)); // TBytes = managed
        Line := Line + ' - (old) ' + TEncoding.ANSI.GetString(OutBuf);
      end;

    end;

    Memo1.Lines.Add(Line);
    SQLQuery1.Next;

  end;
  SQLQuery1.Close;
  SQLite3Connection1.Close;

end;

procedure TForm1.ExtractCookies;
var
  ALocal, AFileName: string;
  Encrypted_Value: TBytes;
  Key: TBytes;
  DataIn: DATA_BLOB;
  DataOut: DATA_BLOB;
  Line: string;
  Nonce: TBytes;
  InBuf: TBytes;
  OutBuf: TBytes;
begin
  Key := GetEncryptedKey(rgBrowser.ItemIndex);
  Memo1.Clear;
  Memo1.ScrollBars := ssVertical;
  Memo1.Lines.Add('Key = ' + TEncoding.ANSI.GetString(Key));

  ALocal := SysUtils.GetEnvironmentVariable('LOCALAPPDATA');
  if ALocal = '' then exit;
  case rgBrowser.ItemIndex of
    0: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'Google\Chrome\User Data\Default\Network\Cookies';
    1: AFileName := IncludeTrailingPathDelimiter(ALocal) + 'BraveSoftware\Brave-Browser\User Data\Default\Network\Cookies';
    else raise Exception.Create('Onbekende browser');
  end;
  if not FileExists(AFileName) then raise Exception.Create('Cookies niet gevonden');

  try
    SQLite3Connection1.Databasename := AFilename;
    SQLQuery1.SQL.Text := 'SELECT host_key, name, value, encrypted_value FROM cookies';
    SQLQuery1.Open;
  except
    on E: Exception do
    begin
      SQLQuery1.Close;
      SQLite3Connection1.Close;
      ShowMessage('Error opening database' + #13 + E.Message);
      exit;
    end;
  end;

  while not SQLQuery1.EOF do
  begin
    Line := '';
    Line := SQLQuery1.FieldByName('host_key').AsString;
    Line := Line + ' ' + SQLQuery1.FieldByName('name').AsString;

    Encrypted_Value := SQLQuery1.FieldByName('encrypted_value').AsBytes;
    // v10 (i.e. 0x763130)
    // of
    // DPAPI = 0x01000000D08C9DDF0115D1118C7A00C04FC297EB

    if Copy(TEncoding.ANSI.GetString(Encrypted_Value), 1, 3) = 'v10' then
    begin
      // v10 (i.e. 0x763130)
      // data = bytes.fromhex('763130...') # the encrypted cookie
      // nonce = data[3:3+12]
      // ciphertext = data[3+12:-16]
      // tag = data[-16:]
      // cipher = AES.new(decrypted_key, AES.MODE_GCM, nonce=nonce)
      // plaintext = cipher.decrypt_and_verify(ciphertext, tag) # the decrypted cookie

      if Length(Encrypted_Value) > 3 + 12 + 16 then
      begin
        SetLength(Nonce, 12);
        SetLength(InBuf, Length(Encrypted_Value) - 15);
        SetLength(OutBuf, Length(InBuf) - 16 { tag});
        Move(Encrypted_Value[3], Nonce[0], 12);
        Move(Encrypted_Value[15], InBuf[0], Length(InBuf));
        if BDecrypt_GCM('AES', InBuf, @OutBuf[0], Key, Nonce) <> 0 then
        begin
          Line := Line + ' - ' + TEncoding.UTF8.GetString(OutBuf);
        end;
      end;

    end
    else
    begin
      // DPAPI = 0x01000000D08C9DDF0115D1118C7A00C04FC297EB

      DataIn.pbData := @Encrypted_Value[0];
      DataIn.cbData := Length(Encrypted_Value);
      if CryptUnprotectData(@DataIn, nil, nil, nil, nil, 0, @DataOut) then
      begin
        SetLength(OutBuf, DataOut.cbData);
        Move(DataOut.pbData^, OutBuf[0], DataOut.cbData);
        LocalFree(cardinal(DataOut.pbData));
        // LocalFree(cardinal(DataIn.pbData)); // TBytes = managed
        Line := Line + ' - (old) ' + TEncoding.ANSI.GetString(OutBuf);
      end;

    end;

    Memo1.Lines.Add(Line);
    SQLQuery1.Next;

  end;
  SQLQuery1.Close;
  SQLite3Connection1.Close;

end;

end.
